"""
Dependency-Check report loader.
"""

import csv
import json
import logging
import re
from datetime import datetime
from os import path
from typing import List

from reportmix.config.property import ConfigProperty
from reportmix.errors import LoadingError
from reportmix.loader import Loader
from reportmix.models import severity
from reportmix.models.issue import Issue
from reportmix.models.project import Project
from reportmix.models.report import Report
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# Configuration properties
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("report_file", "path to the report file", False, "dependency-check-report.csv")
]


class DependencyCheckLoader(Loader):
    """
    Dependency-Check report loader (CSV required, JSON optional).
    """

    def load(self) -> Report:
        """
        Load the Dependency Check report file (CSV required, JSON optional),
        parse it, map vulnerabilities to issues, and return the list.
        :return: Report of vulnerabilities.
        """
        report_file_path = path.realpath(self.config["report_file"])
        if not (report_file_path.endswith(".csv") and path.exists(report_file_path)):
            raise LoadingError("Dependency-Check report ignored (file not found or not *.csv)")

        logging.debug("Loading report %s", report_file_path)

        try:
            # Load the JSON report to extract scan and project info
            json_report_file_path = re.sub(r"\.csv$", ".json", report_file_path)
            scan, project = {}, {}
            if path.exists(json_report_file_path):
                with open(json_report_file_path, "r", encoding="utf8") as json_report_file:
                    json_report = json.load(json_report_file)
                    scan = json_report["scanInfo"]
                    project = json_report["projectInfo"]

            # Load vulnerabilities from the CSV report and map them to issues
            with open(report_file_path, "r", newline='') as report_file:
                report = csv.DictReader(report_file, delimiter=',', quotechar='"')
                issues = []
                for row in report:
                    if "groupID" in project and "artifactID" in project:
                        project_identifier = project["groupID"] + ":" + project["artifactID"]
                    else:
                        project_identifier = row["Project"]
                    issues.append(Issue(
                        ref="",
                        identifier=row["CVE"],
                        name=row["CVE"],
                        type="VULNERABILITY",
                        category=row["CWE"],
                        description=row["Vulnerability"],
                        more="",
                        action="",
                        effort="",
                        analysis_date=datetime.strptime(row["ScanDate"][:24],
                                                        "%a, %d %b %Y %H:%M:%S"),
                        severity=severity.guess(row["CVSSv3_BaseSeverity"]),
                        score=row["CVSSv3"],
                        confidence=row["CPE Confidence"],
                        evidences=int(row["Evidence Count"]),
                        source=row["Source"],
                        source_date=None,
                        url="",
                        tool=Tool(
                            identifier="dependency_check",
                            name="Dependency-Check",
                            version=scan["engineVersion"] if "engineVersion" in scan else ""
                        ),
                        subject=Subject(
                            identifier=row["Identifiers"],
                            name=row["Description"],
                            description=row["DependencyName"],
                            version="",
                            location=row["DependencyPath"],
                            license=row["License"]
                        ),
                        project=Project(
                            identifier=project_identifier,
                            name=row["Project"],
                            version=project["version"] if "version" in project else ""
                        )
                    ))
                tool_version = issues[0].tool.version if len(issues) > 0 else ""
                return Report(issues, [Tool("dependency_check", "Dependency-Check", tool_version)])
        except Exception as ex:
            raise LoadingError("Failed to load, parse and map the report: {}".format(ex))



"""
CSV report exporter.
"""

import csv
from typing import List

from reportmix.exporter import Exporter
from reportmix.models.report import Report


class CsvExporter(Exporter):
    """
    Export a merged report to a CSV file.
    """

    def export(self, report: Report, output_file: str, fields: List[str]):
        with open(output_file, "w", newline='', encoding='utf-8') as file:
            writer = csv.DictWriter(file, fieldnames=fields, extrasaction='ignore',
                                    delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            writer.writeheader()
            for issue in report.issues:
                writer.writerow(issue.flatten())



"""
JSON report exporter.
"""

import json
from typing import List

from reportmix.exporter import Exporter
from reportmix.models.report import Report


class JsonExporter(Exporter):
    """
    Export a merged report to a JSON file.
    """

    def export(self, report: Report, output_file: str, fields: List[str]):
        with open(output_file, "w") as file:
            json.dump([i.to_dict() for i in report.issues], file, default=str)



"""
Report model.
"""
from typing import List

from reportmix.models.issue import Issue
from reportmix.models.tool import Tool


class Report:
    """
    A mixed report of issues found by tools.
    """

    def __init__(self, issues: List[Issue], tools: List[Tool]):
        """
        Initialize a report of issues found by tools.
        :param issues: List of issues (may be empty)
        :param tools: Tools involved in identifying issues
        """
        self.issues = issues
        self.tools = tools

    def extend(self, report: "Report"):
        """
        Extend report by appending issues and tools from another report.
        :param report: Report to append.
        """
        self.issues.extend(report.issues)
        self.tools.extend(report.tools)



"""
ReportMix report loader.
"""

import csv
import logging
from datetime import datetime
from os import path
from typing import List

from reportmix.config.property import ConfigProperty
from reportmix.errors import LoadingError
from reportmix.loader import Loader
from reportmix.models import severity
from reportmix.models.issue import Issue
from reportmix.models.project import Project
from reportmix.models.report import Report
from reportmix.models.subject import Subject
from reportmix.models.tool import Tool

# Configuration properties
PROPERTIES: List[ConfigProperty] = [
    ConfigProperty("report_file", "path to the report file", False)
]


class ReportMixLoader(Loader):
    """
    ReportMix report loader (CSV required).
    """

    def load(self) -> Report:
        """
        Load the ReportMix report file (CSV required), parse it, and return the
        list of issues.
        :return: Loaded issues.
        """
        if "report_file" not in self.config or self.config["report_file"] is None:
            raise LoadingError("ReportMix report ignored (report file path required)")

        report_file_path = path.realpath(self.config["report_file"])
        if not (report_file_path.endswith(".csv") and path.exists(report_file_path)):
            raise LoadingError("ReportMix report ignored (file not found or not *.csv)")

        logging.debug("Loading report %s", report_file_path)

        try:
            # Load issues from the CSV report
            with open(report_file_path, "r", newline='') as report_file:
                report = csv.DictReader(report_file, delimiter=',', quotechar='"')
                issues = []
                for row in report:
                    analysis_date = None
                    if row.get("analysis_date"):
                        analysis_date = datetime.fromisoformat(row["analysis_date"])
                    source_date = None
                    if row.get("source_date"):
                        source_date = datetime.fromisoformat(row["source_date"])
                    evidences = 1
                    if row.get("evidences", "1").isdecimal():
                        evidences = int(row.get("evidences", "1"))
                    issues.append(Issue(
                        ref=row.get("ref"),
                        identifier=row["identifier"],  # Required
                        name=row.get("name"),
                        type=row.get("type"),
                        category=row.get("category"),
                        description=row.get("description"),
                        more=row.get("more"),
                        action=row.get("action"),
                        effort=row.get("effort"),
                        analysis_date=analysis_date,
                        severity=severity.from_identifier(row.get("severity")),
                        score=row.get("score"),
                        confidence=row.get("confidence"),
                        evidences=evidences,
                        source=row.get("source"),
                        source_date=source_date,
                        url=row.get("url"),
                        tool=Tool(
                            identifier=row["tool_identifier"],  # Required
                            name=row.get("tool_name"),
                            version=row.get("tool_version"),
                        ),
                        subject=Subject(
                            identifier=row["subject_identifier"],  # Required
                            name=row.get("subject_name"),
                            description=row.get("subject_description"),
                            version=row.get("subject_version"),
                            location=row.get("subject_location"),
                            license=row.get("subject_license"),
                        ),
                        project=Project(
                            identifier=row.get("project_identifier"),
                            name=row.get("project_name"),
                            version=row.get("project_version"),
                        ),
                        # meta ignored
                        # hash ignored
                    ))
                tools = list({i.tool.identifier: i.tool for i in issues}.values())
                return Report(issues, tools)
        except Exception as ex:
            raise LoadingError("Failed to load and parse the report: {}".format(ex))



"""
Custom exceptions.
"""


class AppError(Exception):
    """
    Generic exception for application errors.
    """


class LoadingError(AppError):
    """
    Exception for report loaders errors.
    """
